// expr.cc
//
// Created 1998 by Karri-Pekka Laakso
//
// $Id: expr.cc,v 1.21 1999/10/22 14:03:09 jks Exp $
//

#ifdef MAKEDEPEND
// makedepend fails to deal with a comment in our copy of cstring
#define __CSTRING__
// The following avoids a spurious #error in Linux' bytesex.h and
// shouldn't cause problems even if you aren't compiling on a 386.
#define __i386__
#endif

#include "expr.h"
#include "variable.h"
#include "record.h"
#include "main.h"
#include <string.h>

extern int yylineno;


//---------------------------------------------------------------------
// constant::printCode
//  The code reprinting function.
//
// 
// --------------------------------------------------------------------

void
constant::printCode()
{
    fprintf(c.file, "%f", val); 
}




//---------------------------------------------------------------------
// arithmeticExpr::arithmeticExpr
//  The class of the +, -, *, % -expressions.
//
// 
// --------------------------------------------------------------------

arithmeticExpr::arithmeticExpr(expr *L, arithmeticOp o, expr *R) :
    l(L), r(R), op(o)
{
    
}


//---------------------------------------------------------------------
// arithmeticExpr::printCode
//  The code reprinting function.
//
// 
// --------------------------------------------------------------------

void
arithmeticExpr::printCode()
{
    l->printCode();
    switch (op) {
    case PLUS: 
	fprintf(c.file, " + ");
	break;
    case MINUS: 
	fprintf(c.file, " - ");
	break;
    case MULT: 
	fprintf(c.file, " * ");
	break;
    case DIV: 
	fprintf(c.file, " / ");
	break;
    default:
	fprintf(c.file, " <unknown> ");
    }
    
    r->printCode();
}


//---------------------------------------------------------------------
// arithmeticExpr::printValue
//  The c++ code printing function.
//
// 
// --------------------------------------------------------------------

void
arithmeticExpr::printValue()
{
    l->printValue();
    switch (op) {
    case PLUS: 
	fprintf(c.file, " + ");
	break;
    case MINUS: 
	fprintf(c.file, " - ");
	break;
    case MULT: 
	fprintf(c.file, " * ");
	break;
    case DIV: 
	fprintf(c.file, " / ");
	break;
    default:
	fprintf(c.file, " <unknown> ");
    }
    
    r->printValue();
}


//---------------------------------------------------------------------
// negationExpr::negationExpr
//  The negation (-x) class.
//
// 
// --------------------------------------------------------------------

negationExpr::negationExpr(expr *R) : r(R)
{
    
}


//---------------------------------------------------------------------
// negationExpr::printCode
//  The code reprinting function.
//
// 
// --------------------------------------------------------------------

void
negationExpr::printCode()
{
    fprintf(c.file, " -");
    r->printCode();
}


//---------------------------------------------------------------------
// negationExpr::printValue
//  The c++ code generating function.
//
// 
// --------------------------------------------------------------------

void
negationExpr::printValue()
{
    fprintf(c.file, " -");
    r->printValue();
}




//---------------------------------------------------------------------
// parenExpr::parenExpr
//  The parenthesis (x) class.
//
// 
// --------------------------------------------------------------------

parenExpr::parenExpr(expr *E) : e(E)
{
    
}


//---------------------------------------------------------------------
// parenExpr::printCode
//  The code reprinting function.
//
// 
// --------------------------------------------------------------------

void
parenExpr::printCode()
{
    fprintf(c.file, "(");
    e->printCode();
    fprintf(c.file, ")");
}


//---------------------------------------------------------------------
// parenExpr::printValue
//  The c++ code generating function.
//
// 
// --------------------------------------------------------------------

void
parenExpr::printValue()
{
    fprintf(c.file, "(");
    e->printValue();
    fprintf(c.file, ")");
}



//---------------------------------------------------------------------
// variableRef::variableRef
//  The variable reference class constructors.
//
// 
// --------------------------------------------------------------------

variableRef::variableRef(char *v) : varName(v), recName(0), dFile(""),
    line(yylineno), var(0), rec(0), local(false), wholeName(0),
    myRecord(c.localRecord), useOldValue(false)
{
    dFile = c.datafile;

    if (c.proposalTargetVariable != 0 &&
	c.proposalTargetVariable->giveName().compare(v) == 0 &&
	c.proposalTargetVariable->giveRecord() == 0)
	{
	    useOldValue = true;
	}
}


variableRef::variableRef(char *r, char *v) : varName(v), recName(r),
    dFile(""), line(yylineno), var(0), rec(0), local(true),
    wholeName(0), myRecord(c.localRecord), useOldValue(false)
{
    dFile = c.datafile;

    if (c.proposalTargetVariable != 0 &&
	c.proposalTargetVariable->giveName().compare(v) == 0 &&
	c.proposalTargetVariable->giveRecord() != 0 &&
	c.proposalTargetVariable->giveRecord()->giveName().compare(r) == 0)
	{
	    useOldValue = true;
	}
}


//---------------------------------------------------------------------
// variableRef::printCode
//  The code reprinting function.
//
// 
// --------------------------------------------------------------------

void
variableRef::printCode()
{
    if (rec != 0)
	fprintf(c.file, "%s.", rec->giveName().c_str());
    fprintf(c.file, "%s", var->giveName().c_str());
}


//---------------------------------------------------------------------
// variableRef::printCodeName
//  Print this variable ref's name in C++.
//
// 
// --------------------------------------------------------------------

void
variableRef::printCodeName()
{
    if (rec != 0)
	fprintf(c.file, "%s->", rec->giveCodeName().c_str());
    fprintf(c.file, "%s", var->giveCodeName().c_str());
}


//---------------------------------------------------------------------
// variableRef::printValue
//  The c++ code generating function.
//
// 
// --------------------------------------------------------------------

void
variableRef::printValue()
{
    if (useOldValue)
	fputs("oldValue", c.file);
    else {
	// Usually, if var->giveRecord() != 0, var either is local to
	// our record or belongs to a referenced record. In either
	// case, we need to prepend "myRecord". However, this could be
	// a global reference; we have a list of those so we can see
	// whether rec is on it.
	// 
	// This is somewhat of an ugly hack. (FIXME)
	if (var->giveRecord() != 0 && c.references->contains(rec) == 0)
	    fprintf(c.file, "myRecord->");
	
	if (rec != 0)
	    fprintf(c.file, "%s->", rec->giveCodeName().c_str());
	fprintf(c.file, "%s->value()", var->giveCodeName().c_str());
    }
}


//---------------------------------------------------------------------
// variableRef::findReferences
//  
//
// 
// --------------------------------------------------------------------

void
variableRef::findReferences()
{
    if (recName == 0) {	// Basic variable reference
	if (myRecord != 0) {
	    // Search for a local variable
	    XlistIterator<variable> i = myRecord->allFields()->giveIterator();
	    for ( ; i.valid(); ++i) 
		if (!((*i)->nameSpace()) && (*i)->giveName().compare(varName) == 0) 
		    {
			var = (*i);
			local = true;
			break;
		    }
	}

	if (var == 0) {
	    // Search for a global variable
	    XlistIterator<variable> j = c.variables->giveIterator();
	    for ( ; j.valid(); ++j) 
		if ((*j)->giveName().compare(varName) == 0) {
		    var = (*j);
		    break;
		}
	}

	// By now the variable should have been found. If it was not, it
	// was undefined, so to prevent further errors we define it.
	if (var == 0) {
	    c.localRecord = myRecord;  // Redefinition check requires this
	    var = new variable(varName);
	    c.localRecord = 0;

	    if (myRecord != 0)
		myRecord->allFields()->append(var);
	    else
		c.variables->append(var);

	    c.reference_error(dFile.c_str(), line, "variable '%s' is undefined", varName);
	}

    } else {	// Record variable reference, "A.B"
	
	// First search for the record reference
	if (myRecord != 0) {
	    // We are inside a record; search local references first
	    XlistIterator<variable> i = myRecord->allFields()->giveIterator();
	    for ( ; i.valid(); ++i) 
		if ( ((*i)->nameSpace() != 0) 
		     && ((*i)->giveName().compare(recName) == 0) ) {
		    rec = (*i);
		    break;
		}
	}
	if (rec == 0) {
	    // Try all global references
	    XlistIterator<variable> i = c.references->giveIterator();
	    for ( ; i.valid(); ++i) 
		if ( ((*i)->nameSpace() != 0) 
		     && ((*i)->giveName().compare(recName) == 0) ) {
		    rec = (*i);
		    local = false;
		    break;
		}
	}

	if (rec == 0) {
	    // The record reference was undefined, so we define it to
	    // prevent some further errors.
	    c.reference_error(dFile.c_str(), line,
			      "record reference '%s' for '%s.%s' is undefined",
			      recName, recName, varName);
	    c.localRecord = myRecord; // Redefinition check requires this

	    rec = new recordRef(recName);
	    if (myRecord != 0)
		myRecord->allFields()->append(rec);
	    else
		c.references->append(rec);

	    // Reference finding has to be done now to ensure that the
	    // reference points to a record
	    rec->findReferences();
	    
	    // Define the variable, too
	    var = new variable(varName);
	    c.localRecord = 0;
	    rec->nameSpace()->append(var);
	    
	} else {
	    // Record reference is ok. Start by doing the reference
	    // finding to ensure, that there exists a record behind
	    // the reference.
	    rec->findReferences();
	    
	    // Search for the variable in the records definitions
	    XlistIterator<variable> i = rec->nameSpace()->giveIterator();
	    for ( ; i.valid(); ++i) 
		if (!((*i)->nameSpace()) && (*i)->giveName().compare(varName) == 0) {
		    var = (*i);
		    break;
		}
	    
	    // If the variable was undefined, we define it to prevent further
	    // errors. 
	    if (var == 0) {
		c.localRecord = myRecord; // Redefinition check requires this
		var = new variable(varName);
		c.localRecord = 0;
		rec->nameSpace()->append(var);
		c.reference_error(dFile.c_str(), line,
				  "no variable '%s' in record '%s'", varName, recName);
	    }
	}
    }

    return;
}


//---------------------------------------------------------------------
// variableRef::collectVars
//  
//
// 
// --------------------------------------------------------------------

void
variableRef::collectVars(Xlist<char> *l)
{
    char *myRecName = "";
    if (local == true && myRecord != 0)
	myRecName = "myRecord->";
	    
    if (wholeName == 0) {
	if (rec != 0) {
	    wholeName = new char[strlen(myRecName) +
				 rec->giveCodeName().length() + strlen("->") +
				 var->giveCodeName().length() + 1];

	    strcpy(wholeName, myRecName);
	    strcat(wholeName, rec->giveCodeName().c_str());
	    strcat(wholeName, "->");
	    strcat(wholeName, var->giveCodeName().c_str());
	} else {
	    wholeName = new char[strlen(myRecName) +
				 var->giveCodeName().length() + 1];

	    strcpy(wholeName, myRecName);
	    strcat(wholeName, var->giveCodeName().c_str());
	}
    }

    l->append(wholeName);
}


//---------------------------------------------------------------------
// functionCall::functionCall
//  The function call (f(x)) class constructor.
//
// 
// --------------------------------------------------------------------

functionCall::functionCall(char *f, Xlist<expr> *l) : func(0), params(l)
{
    // Let's start by counting the parameters
    XlistIterator<expr> j = params->giveIterator();
    int count = 0; 
    
    for ( ; j.valid(); ++j)
	count++;
     
    // Was this function previously defined?
    XlistIterator<function> i = c.functions->giveIterator();
    for ( ; i.valid(); ++i) 
	if ((strcmp(f, (*i)->giveName()) == 0) && (*i)->giveArgs() == count) {
	    func = (*i);
	    break;
	}

    if (func == 0) {
	func = new function(f, count);
	c.functions->append(func);
    }
    
}


//---------------------------------------------------------------------
// functionCall::printCode
//  The code reprinting function.
//
// 
// --------------------------------------------------------------------

void
functionCall::printCode()
{
    fprintf(c.file, "%s(", func->giveName());
    XlistIterator<expr> i = params->giveIterator();
    for ( ; i.valid(); ++i) {
	(*i)->printCode();
	if (!i.atLast())
	    fprintf(c.file, ", ");
    }
    fprintf(c.file, ")");
}


//---------------------------------------------------------------------
// functionCall::printValue
//  The code reprinting function.
//
// 
// --------------------------------------------------------------------

void
functionCall::printValue()
{
    fprintf(c.file, "%s(", func->giveName());
    XlistIterator<expr> i = params->giveIterator();
    for ( ; i.valid(); ++i) {
	(*i)->printValue();
	if (!i.atLast())
	    fprintf(c.file, ", ");
    }
    fprintf(c.file, ")");
}


//---------------------------------------------------------------------
// functionCall::findReferences
//  Find references to variables.
//
// 
// --------------------------------------------------------------------

void
functionCall::findReferences()
{
    XlistIterator<expr> i = params->giveIterator();
    for ( ; i.valid(); ++i)
	(*i)->findReferences();
}


//---------------------------------------------------------------------
// functionCall::collectVars
//  
//
// 
// --------------------------------------------------------------------

void
functionCall::collectVars(Xlist<char> *l)
{
    XlistIterator<expr> i = params->giveIterator();
    for ( ; i.valid(); ++i)
	(*i)->collectVars(l);
}

//----------------------------------------------------------------------
// aggregateFunctionCall::aggregateFunctionCall
//----------------------------------------------------------------------

aggregateFunctionCall::aggregateFunctionCall(char *function,
					     char *record,
					     char *variable = 0):
    aggregateFunctionName(function), 
    argumentRecordName(record), argumentRecord(0),
    argumentVariableName(variable), argumentVariable(0),
    dFile(c.datafile), dLine(yylineno)
{
    // Since records only get created during initialization, the
    // count function is constant throughout the simulation.
    // The others _can_ be constant; as of now, we skip the analysis
    // and just play safe by claiming they aren't. (FIXME)
    if (strcmp(aggregateFunctionName, "count") == 0)
	constant = true;
    else
	constant = false;
}

//----------------------------------------------------------------------
// aggregateFunctionCall::printCode
//----------------------------------------------------------------------

void
aggregateFunctionCall::printCode()
{
    if (argumentVariableName == 0) {
	// Currently, only the `count' function takes only a record as an
	// argument.
	fprintf(c.file, "%s(%s)", 
		aggregateFunctionName, argumentRecordName);
    } else {
	fprintf(c.file, "%s(%s.%s)", 
		aggregateFunctionName, argumentRecordName, 
		argumentVariableName);
    }
}

//----------------------------------------------------------------------
// aggregateFunctionCall::printValue
//----------------------------------------------------------------------

void
aggregateFunctionCall::printValue()
{
    if (strcmp(aggregateFunctionName, "count") == 0)
	fprintf(c.file, "record_%s::count()", argumentRecordName);
    else {
	bool sum = false, prd = false, min = false, max = false, avg = false;

	sum = (strcmp(aggregateFunctionName, "sum") == 0);
	prd = (strcmp(aggregateFunctionName, "product") == 0);
	min = (strcmp(aggregateFunctionName, "min") == 0);
	max = (strcmp(aggregateFunctionName, "max") == 0);
	avg = (strcmp(aggregateFunctionName, "avg") == 0);

	if (sum || prd || min || max) {
	    fprintf(c.file, "record_%s::%s_%s()",
		    argumentRecordName, aggregateFunctionName, 
		    argumentVariable->giveCodeName().c_str());
	} else if (avg) {
	    fprintf(c.file, "(record_%s::sum_%s()/record_%s::count())",
		    argumentRecordName, argumentVariable->giveCodeName().c_str(),
		    argumentRecordName);
	} else {
	    fprintf(stderr, "internal error: invalid aggregate function %s\n",
		    aggregateFunctionName);
	    exit(1);
	}
    }

}

//----------------------------------------------------------------------
// aggregateFunctionCall::printImplementationPrototype
//----------------------------------------------------------------------

void 
aggregateFunctionCall::printImplementationPrototype(string function, variable* var)
{
    if (function == "avg")	// Does not need separate implementation code
	return;

    fprintf(c.file, "static realval %s_%s();\n    ", 
	    function.c_str(), var->giveCodeName().c_str()); }

//----------------------------------------------------------------------
// aggregateFunctionCall::printImplementationCode
//----------------------------------------------------------------------

void 
aggregateFunctionCall::printImplementationCode(string function, variable* var)
{
    if (function == "avg")	// Does not need separate implementation code
	return;
    else if (function == "sum" || function == "product") {
	bool summing = (function == "sum"); // multiplying otherwise
	const char *recordName = var->giveRecord()->giveName().c_str();
	const char *varName = var->giveCodeName().c_str();
	const char *functionName = function.c_str();

	fprintf(c.file, 
		"realval\n"
		"record_%s::%s_%s()\n"
		"{\n"
		"    realval %s = %s;\n"
		"\n"
		"    for(vector<record_%s*>::iterator i = instances->begin();\n"
		"        i != instances->end();\n"
		"        i++)\n"
		"       {\n"
		"           %s %s= (*i)->%s->value();\n"
		"       }\n"
		"\n"
		"    return %s;\n"
		"}\n",
		recordName, functionName, varName,
		functionName, (summing ? "0" : "1"),
		recordName,
		functionName, (summing ? "+" : "*"), varName,
		functionName);

    } else if (function == "min" || function == "max") {
	bool minimizing = (function == "min"); // maximizing otherwise
	const char *recordName = var->giveRecord()->giveName().c_str();
	const char *varName = var->giveCodeName().c_str();
	const char *functionName = function.c_str();
	// name of "accumulator variable":
	const char *accuName = (function + "imum").c_str();

	fprintf(c.file,
		"realval\n"
		"record_%s::%s_%s()\n"
		"{\n"
		"    realval %s;\n"
		"\n"
		"    if (instances->empty()) {\n"
		"        %s = %sINFINITY;\n" // the only reasonable thing to return
		"    } else {\n"
		"        %s = (*instances)[0]->%s->value();\n"
		"        for (vector<record_%s*>::iterator i = instances->begin() + 1;\n"
		"             i != instances->end();\n"
		"             i++)\n"
		"            {\n"
		"                %s = %s(%s, (*i)->%s->value());\n"
		"            }\n"
		"    }\n"
		"\n"
		"    return %s;\n"
		"}\n",
		recordName, functionName, varName,
		accuName,
		accuName, (minimizing ? "+" : "-"),
		accuName, varName,
		recordName,
		accuName, functionName, accuName, varName,
		accuName);
	
    } else {
	outputDiagnostic("internal error: unknown aggregate function %s",
			 function.c_str());
    }
}


//----------------------------------------------------------------------
// aggregateFunctionCall::findReferences
//----------------------------------------------------------------------

void
aggregateFunctionCall::findReferences()
{
    // FIXME: We ought to understand local references, i.e.
    // record foo { var x ~ norm(0,1); var y ~ norm(aggr_sum(x), aggr_max(x));
    //
    // if (myRecord != 0) {
    // }

    //fprintf(stderr, "aFC::fR(): %s %s %s\n", aggregateFunctionName, argumentRecordName, argumentVariableName);

    {
	XlistIterator<record> i = c.records->giveIterator();
	for (i.first(); i.valid(); i++) {
	    if ((*i)->giveName().compare(argumentRecordName) == 0) {
		argumentRecord = *i;
		break;
	    }
	}
	if (argumentRecord == 0) {
	    c.reference_error(dFile.c_str(), dLine, 
			      "no such record: %s", argumentRecordName);
	    return;
	}
	argumentRecord->findReferences();
    }

    if (argumentVariableName != 0) {
	//fprintf(stderr, "-- going on...\n");
	XlistIterator<variable> i = argumentRecord->allFields()->giveIterator();
	for (i.first(); i.valid(); i++) {
	    if ((*i)->giveName().compare(argumentVariableName) == 0) {
		argumentVariable = *i;
	    }
	}
	if (argumentVariable == 0) {
	    c.reference_error(dFile.c_str(), dLine, 
			      "no such variable in record %s: %s",
			      argumentRecordName, argumentVariableName);
	    return;
	}

	// In the current implementation, variables that are
	// referenced in aggregate function calls keep track of the
	// functions and then call the printImplementation*()
	// functions to make sure the needed implementations are
	// included.

	// FIXME: This may not be the obvious place for this call, but
	// it is the simplest. Also, the information that "avg" uses
	// "sum" is spread over too many places.
	if (strcmp(aggregateFunctionName, "avg") == 0)
	    argumentVariable->registerAggregateUse("sum");
	else
	    argumentVariable->registerAggregateUse(aggregateFunctionName);
    }

}

//----------------------------------------------------------------------
// aggregateFunctionCall::collectVars
//
// The way this function works is somewhat of a crude hack (FIXME).
// The way of adding dependencies doesn't generalize very well to the
// case of aggregate functions: collectVars gets called to gather a
// list of variable names such as "myRecord->var_x", "global_var_y",
// "myRecord->some_reference->var_z", and code is generated to call
// addDependents for everything in the list.
//
// Now we would like to return a list of all variables of a given
// type, which we obviously can't do at compile time. Here's the hack:
// There's a static function dependencyAdder that returns a pointer to
// an object whose sole purpose is to receive addDependency() calls
// and pass them on to a static list of variable instances. That way,
// the caller of collectVars can happily add ->addDependency() to the 
// string we return.
//
// I feel slightly disgusted at this hack, and I'm pretty sure there's
// a better way to solve this that involves changing the dependency
// generation mechanism in a fundamental way. (Perhaps the Visitor
// design pattern is the Right Thing?)
//----------------------------------------------------------------------

void
aggregateFunctionCall::collectVars(Xlist<char> *l)
{
    if (argumentVariableName == 0)
	return;			// Doesn't refer to a variable => no dependencies.

    if (argumentVariable == 0 || argumentRecord == 0)
	findReferences();

    string r = argumentVariable->giveType();
    r += "::dependencyAdder()";

    // This probably works:
    //l->append(const_cast<char*>(r.c_str()));

    // ...but to avoid the "probably" bit above:
    char *tmp = new char[r.length()+1];
    strcpy(tmp, r.c_str());
    l->append(tmp);
    // This leaks memory, so a better answer is to collect C++ strings
    // instead of char*:s. Even better is to get rid of this hack.
}

//-----------------------------//
// Local Variables:            //
// mode: c++                   //
// make-backup-files: t        //
// c-basic-offset: 4           //
// End:                        //
//-----------------------------//
