// main.cc
//
// Created 1998 Karri-Pekka Laakso
//
// $Id: main.cc,v 1.19 1999/11/01 09:53:33 jks Exp $
//

#include "main.h"

#include <vector>
#include <iterator>
#include <algorithm>
#include <functional>
#include <string>
#include <cstring>
#include <ctime>
#include <cstdio>
#include <cstdlib>
using namespace std;

#include <signal.h>
#include <unistd.h>
#if HAVE_FNMATCH
#include <fnmatch.h>
#else
#error You do not have a working fnmatch function. 
#error Please report a bug in Bassist: 
#error it should include an fnmatch implementation for you.
// FIXME
#endif

#include "common.h"
#include "composite.h"
#include "parameter.h"

//----------------------------------------
// delayedOption
//----------------------------------------

void delayedIEOption::execute(void)
{
    if (_patterns.length() == 0)
	return;

    // Make a copy of the pattern, since we use strtok which modifies it.
    char *patternCopy = new char[_patterns.length() + 1];
    _patterns.copy(patternCopy, string::npos);
    patternCopy[_patterns.length()] = 0;
    
    // FIXME: strtok is not in accordance with The C++ Way (and will bite
    // if we need to go multithreaded or anything)
    for (char *currentPattern = strtok(patternCopy, ",");
	 currentPattern != 0;
	 currentPattern = strtok(0, ",")) 
	{
	    vector<parameter*>::const_iterator i = parameter::begin();
	    while (i != parameter::end()) {
		if (fnmatch(currentPattern, (*i)->identifier().c_str(), 0) == 0) {
		    (*i)->printOnOutput(_include);
		}
		i++;
	    }

	    vector<composite*>::const_iterator j = composite::begin();
	    while (j != composite::end()) {
		if (fnmatch(currentPattern, (*j)->identifier().c_str(), 0) == 0) {
		    (*j)->printOnOutput(_include);
		}
		j++;
	    }
	}
    
    delete[] patternCopy;
}

void delayedDataExcludeOption::execute(void)
{
    for (vector<parameter*>::const_iterator i = parameter::begin();
	 i != parameter::end();
	 i++)
	{
	    if ((*i)->isData())
		(*i)->printOnOutput(false);
	}

    for (vector<composite*>::const_iterator i = composite::begin();
	 i!= composite::end();
	 i++)
	{
	    (*i)->printOnOutput(false);
	}
}

//----------------------------------------
// Simulator
//----------------------------------------



//---------------------------------------------------------------------
// simulator::simulator
//  The simulator constructor.
//
// --------------------------------------------------------------------

simulator::simulator(int ac, char **av) :
    argc(ac), argv(av), stdoutBuffer(0), progName(av[0]), iniFile(0), 
    randSeed(0), nrOfIterations(0), skipIterations(0), everyNth(1), 
    iteration(0), tmpResultRequested(false), lastModified(0)
{
    delayedOptions = new vector<delayedOption*>;
    // Initialize the random seed number from time
    randSeed = (unsigned int) time(0); 

}

//---------------------------------------------------------------------
// simulator::~simulator
//  The simulator destructor.
//
// --------------------------------------------------------------------

simulator::~simulator()
{
    // stdoutBuffer is allocated with new; delete[] it
    setBuffering(false, 0);
}

void
simulator::setBuffering(bool lineBuffering, size_t size = BUFSIZ)
{
    char* oldBuffer = stdoutBuffer;

    if (size == 0) {
	// unbuffered
	stdoutBuffer = 0;	// see note near end of function
	if (setvbuf(stdout, 0, _IONBF, 0) != 0) {
	    fprintf(stderr, "Cannot unset stdout buffer: %s\n",
		    strerror(errno));
	}
    } else {
	// line or fully buffered
	stdoutBuffer = new char[size];
	if(setvbuf(stdout, stdoutBuffer, 
		   (lineBuffering? _IOLBF : _IOFBF), size) 
	   != 0) 
	    {
		fprintf(stderr, 
			"Cannot set stdout %s buffered with size %d: %s\n",
			lineBuffering? "line" : "fully", 
			size,
			strerror(errno));
		exit(1);
	    }
    }

    // If there was an old buffer, get rid of it. 

    // Note that the following invariant must hold between calls to
    // this function: stdoutBuffer != 0 if and only if stdoutBuffer
    // points to something allocated with new[].

    if (oldBuffer != 0) {
	delete[] oldBuffer;
    }	
}

//---------------------------------------------------------------------
// simulator::usage
//  The usage message.
//
// --------------------------------------------------------------------

void
simulator::usage()
{
    printModelName(stderr);
    fprintf(stderr, "
Usage: %s -h -m -D [-d debug_flags] [-r <randseed>] <iterations>
    [-s <iterations>] [-n <n>] [[-I <patterns> | -E <patterns> | -S] ...]
    [-L[<bufsize>] | -B[<bufsize>] ] <datafile> ... [-i inifile] 

-h                this help message
-m                print model definition and quit
-D                print model data example and quit
-r <randseed>     set random seed number to <randseed>
-s <iterations>   skip printing of first <iterations>
-n <n>            print every <n>'th iteration
-I <patterns>     include variables in output (see manual page)
-E <patterns>     exclude variables from output (see manual page)
-S                exclude non-simulated variables (data & composites)
-L[<bufsize>]     make output line buffered with optional buffer size
-B[<bufsize>]     make output fully buffered with optional buffer size

Current debug flags:
qQ - quantities
pP - parametres
dD - distributions
rR - proposers
iI - initialisation phase
sS - simulation
mM - Metropolis-Hastings function
oO - a special debug category for Kari Vasko
+ - all
", progName);

    exit(1);
}


//---------------------------------------------------------------------
// simulator::readOptions
//  Read the command line options. 
//
// --------------------------------------------------------------------

void
simulator::readOptions()
{
    string debugArgs = "";
   
    int opt;
    bool seenIE = false;
    
    while ((opt = getopt(argc, argv, "hmDd:r:s:n:i:I:E:SB::L::")) != -1) {
	switch (opt) {
	case 'h':
	    usage();
	    break; // Not reached

	case 'm':
	    printModelDef(stdout);
	    exit(0);
	    break; // Not reached
	    
	case 'D':
	    printDataExample(stdout);
	    exit(0);
	    break; // Not reached
	    
	case 'd':  
	    debugArgs = optarg;
	    break;

	case 'r':
	    randSeed = (unsigned int) atoi(optarg);
	    break;

	case 's':
	    skipIterations = atoi(optarg);
	    break;

	case 'n':
	    everyNth = atoi(optarg);
	    break;

	case 'i':
	    iniFile = optarg;
	    break;

	case 'I':
	    if (!seenIE) {
		// This is the first '-I' or '-E': do implicit '-E *' first
		delayedOptions->push_back(new delayedExcludeOption("*"));
	    }
	    delayedOptions->push_back(new delayedIncludeOption(optarg));
	    seenIE = true;
	    break;

	case 'E':
	    if (!seenIE) {
		// This is the first '-I' or '-E': do implicit '-I *' first
		delayedOptions->push_back(new delayedIncludeOption("*"));
	    }
	    delayedOptions->push_back(new delayedExcludeOption(optarg));
	    seenIE = true;
	    break;

	case 'S':
	    if (!seenIE) {
		// Implicit '-I *'
		delayedOptions->push_back(new delayedIncludeOption("*"));
	    }
	    delayedOptions->push_back(new delayedDataExcludeOption());
	    seenIE = true;
	    break;

	case 'B':
	case 'L':
	    {
		int bufferSize;
		bool lineBuffering;
		
		if (opt == 'B')
		    lineBuffering = false;
		else
		lineBuffering = true;
		
		if (optarg) {
		    if (sscanf(optarg, "%d", &bufferSize) != 1) {
			fprintf(stderr, "Error: option -%c expects a numeric argument, got %s\n\n",
				opt, optarg);
			usage();
		    }
		    setBuffering(lineBuffering, bufferSize);
		} else {
		    setBuffering(lineBuffering);
		}

		break;
	    }

	default:
	    usage();
	}
    }

    if (!seenIE) {
	// No '-I' or '-E' options: implicit '-I * -S'
	delayedOptions->push_back(new delayedIncludeOption("*"));
	delayedOptions->push_back(new delayedDataExcludeOption());
    }

    if (argc - optind < 2) {
	if (argc - optind < 1) {
	    fprintf(stderr, "Error: missing iterations and datafile name(s)\n");
	} else {
	    fprintf(stderr, "Error: missing datafile name(s)\n");
	}
	usage();
    }

    nrOfIterations = atoi(argv[optind]);
    if (nrOfIterations < skipIterations) {
	fprintf(stderr, "Error: number of iterations must be greater or equal than print start\n");
	exit(1);
    }
    
    debugInit(debugArgs);
}

//---------------------------------------------------------------------
// simulator::executeDelayedOptions
//---------------------------------------------------------------------

void 
simulator::executeDelayedOptions(void)
{
    for_each(delayedOptions->begin(), delayedOptions->end(), 
	     mem_fun(&delayedOption::execute));

    // It's apparently difficult to use operator delete with for_each.
    for(vector<delayedOption *>::const_iterator i = delayedOptions->begin();
	i != delayedOptions->end();
	i++) 
	{
	    delete *i;
	}
    
    delete delayedOptions;
    delayedOptions = 0;
}

//---------------------------------------------------------------------
// simulator::printCommandLine
//  Print the command line as it was.
//
// Printing the command line gives the user at least some chance to
// remember, where the results of this simulation came from.
// --------------------------------------------------------------------

void
simulator::printCommandLine()
{
    int i;
    time_t t = time(0);

    printf("#");
    for(i = 0; i < argc; i++) 
	printf(" %s", argv[i]);
    printf("\n");

    printf("# %s", asctime(localtime(&t)));
}



//---------------------------------------------------------------------
// simulator::printHeader
//  Print the data file header for Javist and other tools.
//
// A check for proper parameter initialization is done here, too.
// --------------------------------------------------------------------

void
simulator::printHeader()
{
    // The start of the header; misc stuff
    printf("# <header>\n");
    printModelName(stdout);
    printCommandLine();
    printf("# Random generator seed set to %u\n", randSeed); 

    long unsigned counter = 0;

    // Optional DEBUG: list all quantities is simulation
    if (debugIsEnabled('s')) {
	for (vector<quantity*>::const_iterator i = quantity::begin();
	     i != quantity::end();
	     i++)
	    {
		quantity *q = *i;
		counter++;
		DEBUG('S', "%s: %G\n", q->identifier().c_str(), q->value());
	    }
    }
    DEBUG('s', "Quantities in simulation: %lu\n", counter);

    counter = 0;

    parameter *p;

    // List all simulation attributes, whose values will be printed
    // during simulation, in the same order
    
    printf("# <attributes>");

    // First simulatable parameters
    for (vector<parameter*>::const_iterator i = parameter::begin(); 
	 i != parameter::end(); 
	 i++) 
	{
	    p = *i;
	    counter++;
	    if (p->printOnOutput()) {
		p->printName(stdout);
	    }
	    
	    // In this same pass, check that every parameter was initialized properly.
	    if (! p->isInited()) {
		fprintf(stderr, "Error: Uninitialized parameter");
		p->printName(stderr);
		fprintf(stderr, " in simulation.\n");
		return;
	    }
	}

    // Then the deterministic values 
    vector<composite*>::const_iterator j = composite::begin();
    while (j != composite::end()) {
	if ((*j)->printOnOutput())
	    (*j)->printName(stdout);
	j++;
    }
    
    printf(" </attributes>\n");

    DEBUG('s', "# Simulatable parameters in simulation: %lu\n", counter);

    printf("# <record separator> '\\n' </record separator>\n");
    printf("# <field separator> ' ' </field separator>\n");
    printf("# </header>\n");
    
    if (counter == 0) {
	printf("# Nothing to simulate, exiting\n\n");
	exit(0);
    }

    // Make sure this print really goes somewhere
    fflush(stdout);
}



//---------------------------------------------------------------------
// simulator::start
//  The simulator main function.
//
// All the simulation initialization and data reading is called from
// main. simulate() is the place, where everything happens during
// simulation. 
// --------------------------------------------------------------------

int
simulator::start()
{
    readOptions();
    
    // Set random generator
    rng.setSeed(randSeed);

    createGlobals();

    // Read data files
    for (int opt = optind+1 ; opt < argc; opt++) 
	readInput(argv[opt], false);

    // Read optional initilization file
    if (iniFile != 0)
	readInput(iniFile, true);
    
    // Initialize
    initGlobals();
    initRecords();

    // Handle delayed options
    executeDelayedOptions();

    // Start simulation
    printHeader();
    simulate();

    return 0;
}



//---------------------------------------------------------------------
// simulator::printTmpResult
//  Prints the temporary results for the iteration.
//
// This function is called, when user has requested a temporay result
// by sending a SIGUSR1 signal. The default output file is called
// "tmpresult". 
// --------------------------------------------------------------------

void
simulator::printTmpResult()
{
    FILE *file = fopen("tmpresult", "w");
	    
    if (file == 0)
	file = stderr;	// Print to stderr, if cannot open file

    fprintf(file, "\nLoop %d\n\n", iteration);

    for (vector<parameter*>::const_iterator i = parameter::begin(); 
	 i != parameter::end(); 
	 i++) 
	{
	    parameter *p = *i;
	    fprintf(file, "%s: %G\n", p->identifier().c_str(), p->value());
	}
	    
    fprintf(file, "\n");
	    
    if (file != stderr)
	fclose(file);
}


//---------------------------------------------------------------------
// simulator::simulate
//  The actual simulation function.
//
// Every simulatable parameter's value is updated in the MH
// function. If the values of this iteration should be printed, they
// are. If the user requested a temporary result by sending SIGUSR1
// (tmpResultRequested is set), such is printed for the current
// iteration.
// --------------------------------------------------------------------

void
simulator::simulate()
{  
    // Set the signals
    signal(SIGTERM, sigTermHandler);
    signal(SIGUSR1, sigUsr1Handler);
    
    DEBUG('s', "# Starting simulation...\n");

    bool printIteration = false;
    
    // Indicator for a valid iteration loop. This will become false,
    // if both the old and the new probability in the MH-update of
    // some parameter are < -INFINITY, which indicates an infeasible
    // state of the model.
    bool validIteration = true;
    int invalidCount = 0;
    
    vector<parameter*>::const_iterator params;
    vector<composite*>::const_iterator compos;

    // For each iteration
    for (iteration = 1; iteration <= nrOfIterations; ) {
	DEBUG('S', "# iteration %i:\n", iteration);

	validIteration = true;

	if (iteration > skipIterations && ((iteration - skipIterations) % everyNth) == 0)
	    printIteration = true;
	else
	    printIteration = false;
	
	// For each parameter in the model
	for (params = parameter::begin(); params != parameter::end(); ++params) {
	    validIteration &= (*params)->MH(); // Update value
	    lastModified = (*params);
	}

	if (validIteration) {
	    if (printIteration) {
		// For each parameter and user defined composite
		for (params = parameter::begin(); 
		     params != parameter::end(); 
		     ++params)
		    {
			if ((*params)->printOnOutput())
			    (*params)->printValue(stdout);
		    }
		for (compos = composite::begin(); 
		     compos != composite::end(); 
		     ++compos)
		    {
			if ((*compos)->printOnOutput())
			    (*compos)->printValue(stdout);
		    }
	    
		printf("\n");

		if (tmpResultRequested) {
		    printTmpResult();
		    tmpResultRequested = false;
		}
	    }
	    
	    ++iteration;
	    invalidCount = 0;
	} else {	// Invalid iteration
	    ++invalidCount;
	    if (invalidCount >= 200) {
		fprintf(stderr, "Error: Model state was infeasible for over 200 iterations.
Problem may be due to bad initial values of parameters.
Explicite initial values should be used or improved.
Simulation aborted.\n");
		exit(1);
	    }
	}
    }

    return;
}



//---------------------------------------------------------------------
// simulator::sigTermHandler
//  Signal handler for SIGTERM
//
// This makes sure, that at least the last values get to be saved.
// --------------------------------------------------------------------

void
simulator::sigTermHandler(int)
{
    simulator *s = theInstance();

    printf("\nCaught TERM signal during loop %d, \"%s\" was last changed\n",
	   s->iteration, s->lastModified->identifier().c_str());

    parameter *p;

    for (vector<parameter*>::const_iterator i = parameter::begin(); 
	 i != parameter::end();
	 ++i) 
	{
	    p = *i;
	    printf("%s: %G\n", p->identifier().c_str(), p->value());
	}
    
    exit(0);
}



//---------------------------------------------------------------------
// simulator::sigUsr1Handler
//  The signal handler for SIGUSR1
//
// SIGUSR1 is a way for to get information on the current state of the 
// simulation. By sending the simulation program a SIGUSR1, the
// simulation program writes a file called 'tmpresult' to its working
// directory. It contains the current state of all simulatable
// variables and the iteration count.
// --------------------------------------------------------------------

void
simulator::sigUsr1Handler(int)
{
    simulator *s = theInstance();

    signal(SIGUSR1, sigUsr1Handler);	// We want to trap the next USR1, too. 
    s->tmpResultRequested = true;
    return;
}

// We use the Singleton pattern here.

simulator *simulator::_instance = 0;

// --------------------------------------------------------------------
// theInstance
//  Returns the single instance of this object 
// --------------------------------------------------------------------

inline simulator *
simulator::theInstance()
{
    if (_instance == 0) {
	_instance = new simulator(0, 0);
	// _Or_ we could flag an error, since obviously main() has not
	// supplied the necessary data.
    }

    return _instance;
}

// --------------------------------------------------------------------
// supplyData
//  main() should call this function so as to supply data for the
//  single instance of the object.
// --------------------------------------------------------------------

void
simulator::supplyData(int argc, char **argv)
{
    // If _instance != 0, the object has been constructed prior to this
    // function call, and it is too late to supply the data.
    ASSERT(_instance == 0);
    
    // We now construct the object. Another approach would be to register
    // the supplied data somewhere and call the constructor only in
    // theInstance().
    _instance = new simulator(argc, argv);
}

// Local Variables:
// mode: c++
// make-backup-files: t
// c-basic-offset: 4
// case-replace: nil
// End:

