// scanner.cc
//
// Created 1997 by Karri-Pekka Laakso
//
// $Id: scanner.cc,v 1.4 1999/08/03 15:11:12 jks Exp $
//

#include "scanner.h"
#include <ctype.h>
#include <string.h>
#include "parameter.h"

char scanner::bufReadFormat[20] = { '\0' };
char scanner::idReadFormat[20] = { '\0' };

//---------------------------------------------------------------------
// scanner::scanner
//  
//
// 
// --------------------------------------------------------------------

scanner::scanner(char *fname)
    : ind(0), longLine(false), currentLine(1), eofReached(false)
{
    DEBUG('f', "Constructing scanner for file %s\n", fname);
    buffer[0] = buffer[BUFSIZE] = '\0';

    if (bufReadFormat[0] == '\0') {
	sprintf(bufReadFormat, "%%%d[^\n]", BUFSIZE);
	sprintf(idReadFormat, "%%%d[a-zA-Z0-9_]", MAXID+1);
    }
    
    filename = new char[strlen(fname)+1];
    strcpy(filename, fname);
    
    file = fopen(filename, "r");
    if (file == NULL) {
	fprintf(stderr, "File open error with %s: ", filename);
	perror("");
	exit(1);
    }
}


//---------------------------------------------------------------------
// scanner::fillBuffer
//  Eats whitespace from input buffer and reads in a new line, if end
//  of buffer was reached.
//
// Returns false if end of file was reached, otherwise true.
// --------------------------------------------------------------------

bool
scanner::fillBuffer()
{
    // Eat whitespace
    while (ind < BUFSIZE && isspace(buffer[ind]))
	ind++;

    bool commentRow = false, possibleCommentLine = false;
    
    // Fill if we are at the end of buffer, at a comment row or 5/6
    // done of a long line
    while (buffer[ind] == '\0' || ind == BUFSIZE || commentRow == true ||
	   (longLine && ind > BUFSIZE*5/6))
    {
	char *format;
	int bstart;

	DEBUG('f', "sc.fB: line = %d, ind=%d, blen = %d, buffer=\"%s\"\n",
	      currentLine, ind, strlen(buffer), buffer);
    
	// Are we going to fill or to read a new line?
	if (longLine && (ind > BUFSIZE*5/6)) {
	    strcpy(&buffer[0], &buffer[ind]);
	    // 'ind' chars have been read, so there is space for 'ind' new chars 
	    sprintf(bufFillFormat, "%%%d[^\n]", ind);
	    format = bufFillFormat;
	    bstart = BUFSIZE-ind;
	    DEBUG('f', "sc.fB: fill, ind=%d, bstart=%d\n", ind, bstart);
	} else {
	    DEBUG('f', "sc.fB: new line\n");
	    format = bufReadFormat;
	    bstart = 0;
	    buffer[0]='\0';
	}
	
	// Whatever happens, reading starts next time from the
	// beginning of buffer 
	ind=0;
	
	int res = fscanf(file, format, &buffer[bstart]);
	DEBUG('f', "sc.fB: read result = %d\n", res);
	while (res == 0) {
	    // Line consists of \n only == end of line reached
	    // Eat \n, do all new line stuff and try again
	    fscanf(file, "%*c");

	    currentLine++;
	    possibleCommentLine = true;
	    commentRow = false;
	    
	    // If this was a fill, end of line has already been
	    // reached, so the buffer must be first read to its end
	    if (format == bufFillFormat)
		break;

	    res = fscanf(file, bufReadFormat, &buffer[0]);
	    DEBUG('f', "sc.fB: read result = %d\n", res);
	}
	
	// Handle errors
	if (res < 0) {
	    // End of file reached?
	    if (feof(file)) {
		if (bstart == 0) {
		    eofReached = true;
		    return false;
		}
		// If bstart > 0, this was a buffer fill, and we'll
		// fall through this time, because there's still
		// something to read.
	    } else {
		fprintf(stderr, "Error reading %s:%d: ", filename, currentLine);
		perror("");
		exit(1);
	    }
	}

	int len = strlen(&buffer[0]);

	// Ok, now we have an input line, which has at least one
	// character. 
	
	// Long lines need special handling
	if (len == BUFSIZE) {
	    longLine = true;
	} else {
	    longLine = false;
	}

	// Ignore comment rows
	if (commentRow == true)
	    continue;
	
	// Is this a comment row = a row beginning with '#'?
	if (possibleCommentLine == true && buffer[0] == '#') {
	    commentRow = true;
	    continue;
	}

	possibleCommentLine = false;
	
	// Eat whitespace from the beginning of line
	while (ind < BUFSIZE && isspace(buffer[ind]))
	    ind++;
    }

    return true;
}


//---------------------------------------------------------------------
// scanner::get
//  Requests the argument string from the input line.
//
// If the requested string was not found, exits the program.
// --------------------------------------------------------------------

void
scanner::get(char *string)
{
    if (!ask(string)) {
	fprintf(stderr, "Error in %s:%d: expecting \"%s\"\n", filename,
	       currentLine, string);
	exit(1);
    }
}


//---------------------------------------------------------------------
// scanner::ask
//  Checks if the argument string is in the input.
//
// If the argument string was found, moves the input forward and
// returns true; if it was not found, no move is done and returns
// false. Whitespace is removed from the beginning and the end of the
// argument string.
// --------------------------------------------------------------------

bool
scanner::ask(char *string)
{
    int len = strlen(string);
    char *tempStr, *validStr;

    DEBUG('f', "sc.ask(%s), buffer[%d] =\"%s\"\n", string,
	  ind, &buffer[ind]);
    
    // Whitespace padded (padding from the beginning or the end of
    // string) reads will always fail, because whitespace is 
    // eaten automatically from the buffer. Therefore the whitespace
    // is eaten away also from the string. 
    int i, j;
    for (i = 0; i < len && isspace(string[i]); i++);
    for (j = len; j > i && isspace(string[j-1]); j--);

    // A copy is made of the string, if it ends with whitespace,
    // because the string might be constant.
    if (j != len) {
	tempStr = new char[j-i+1];
	strcpy(tempStr, string);
	validStr = tempStr;
    } else {
	validStr = &string[i];
    }
    
    len = j-i;

    if (!fillBuffer() || (ind+len) > BUFSIZE) {
	// EOF reached or string is longer than the remaining buffer
	return false;
    }
    
    if ((strncmp(validStr, &buffer[ind], len) == 0)) {
	ind += len;
	return true;
    }

    return false;
}


//---------------------------------------------------------------------
// scanner::getId
//  Copies the next identifier from the input stream.
//
// The argument buffer must have space for at least scanner::MAXID
// chars. An identifier may consist of chars a-z, A-Z, 0-9 and _; see
// idReadFormat in constructor. Returns true, if an identifier was
// found, otherwise false. Moves the input stream forward.
// --------------------------------------------------------------------

bool
scanner::getId(char *string)
{
    if (next() != 0) {
	strcpy(string, s);
	return true;
    }

    // No identifier found
    string[0] = '\0';
    return false;
}

//---------------------------------------------------------------------
// scanner::getDoubleValue
//  Tries to read a double from the input stream and return its
//  value. 
//
// If read fails, this function exits. Moves the input stream forward.
// --------------------------------------------------------------------

double
scanner::getDoubleValue()
{
    double value;

    if(!fillBuffer()) {
	fprintf(stderr,
		"Error in %s:%d: expected a double, but reached end of file.\n",
	        filename, currentLine);
	exit(1);
    }
    
    int len;

    if (sscanf(&buffer[ind], "%lf%n", &value, &len) != 1) {
	fprintf(stderr,
		"Error in %s:%d: expected a double, but found \"%s\"\n",
	        filename, currentLine, &buffer[ind]);
	exit(1);
    }

    ind += len;
    return value;
}

//---------------------------------------------------------------------
// scanner::askParameter
//  Assigns the given parameter a value and sets it as data, if the
//  string "name = value" is found from the input.
//
// Name does not have to be the parameters name. The value is of type
// double. Returns true if the sequence was found, false if it was not 
// found. Moves the input forward.
// --------------------------------------------------------------------

// bool
// scanner::askParameter(parameter *p, char *name)
// {
//     double value;

//     // Get the next identifier, but don't move input, in case it
//     // wasn't the one we're looking for
//     char *nextId = next(false);
    
//     if (nextId != 0 && strcmp(name, nextId) == 0) {
// 	// It matched, so first eat the id name
// 	next();
// 	get("=");
// 	value = getDoubleValue();
// 	p->setData(value);
// 	return true;
//     }

//     return false;
// }


// bool
// scanner::askObservationStatus(int *p, char *name)
// {   
//     double value;

//     if (ask(name)) {
// 	get("=");
// 	value = getDoubleValue();
// 	*p = (int)value;
// 	return true;
//     }

//     return false;
// }





//---------------------------------------------------------------------
// scanner::next
//  Returns a pointer to the next identifier in the input stream.
//
// Moves the input forward. Subsequent calls will overwrite the
// contents of the returned pointer. Returns NULL when the input
// stream has ended.  
// --------------------------------------------------------------------

char *
scanner::next(bool moveForward)
{
    DEBUG('f', "sc.next(), buffer[%d] =\"%s\"\n",
	  ind, &buffer[ind]);
    
    if(!fillBuffer()) {
	return 0;
    }
    
    int res = sscanf(&buffer[ind], idReadFormat, s);

    if (res < 1) {	// Read failed
	if (res == 0)
	    return 0;
	else {
	    fprintf(stderr, "Error reading %s:%d:", filename, currentLine);
	    perror("");
	    exit(1);
	}
    }

    if (strlen(s) > MAXID) {
	fprintf(stderr, 
		"Error in %s:%d: identifier exceeds %d characters: \"%s\"\n",
	        filename, currentLine, MAXID, &buffer[ind]);
	exit(1);
    }
	
    if (moveForward)
	ind += strlen(s);
    
    return s;
}


//---------------------------------------------------------------------
// scanner::~scanner
//  Deallocates the scanner.
//
// Closes also the input file.
// --------------------------------------------------------------------

scanner::~scanner()
{
    delete[] filename;
    fclose(file);
}


//////////////////////////////////////////////////
// Local Variables:				//
// mode: c++					//
// make-backup-files: t				//
// End:						//
//////////////////////////////////////////////////
