// list.h
//
// Created 1997 by Karri-Pekka Laakso
//
// $Id: list.h,v 1.4 1999/04/23 10:10:25 jks Exp $
//

#ifndef LIST_H
#define LIST_H

#include "common.h"
#define BASSIST_LIST_REFCOUNT

// Forward declarations
template<class T> class Xlist;
template<class T> class XlistIterator;


/**
  	Class listElement. Instance of the class list
  	consists of instances of the class listElement.

 	The actual items are saved as pointers, that is no copy of the original
 	item is made. This may cause trouble with deallocated values, but
 	this is an organizational structure, not a storage structure.
 
 	The linking is two-way. A reference count is kept in the
 	attribute "count", so that the elements may won't be deleted 
 	from the list, if someone is still pointing it.
 */
template<class T> class XlistElement {
    friend class Xlist<T>;
    friend class XlistIterator<T>;

protected:
/**
   Pointer to item of this listElement.
*/
    T *item;
/// 	Previous element.
    XlistElement<T> *prev;
/// 	Next element.
    XlistElement<T> *next;
/**
   Reference count of this listElement (= number of pointers pointing 
   this listElement
*/
    int count;
    
public:
///
    XlistElement() : item(0), prev(0), next(0), count(0) {};
};

/**
  	Class list is used to maintain various kinds of lists.
  	In Bassist this class is used to store quantities
  	of different kinds.

  	The list is two way linked.
  
  	A reference count is kept on each list element. An element gets
  	deleted only when its reference count drops to zero. If it is
  	remove()d from the list before this, it is moved to a temporary
  	storage array, until its reference count goes to zero. The next
  	and previous links of the moved element are kept sane, though it's
  	not on the list.
  
  	This list is meant for saving pointers to items. No copies of the
  	items are made, and no validity checks are done. This is a pure
  	organizational list, i.e. not a storage list.
  
  	In case a storage list is wanted, use the STL list implementation.
  
	@author Karri-Pekka Laakso
  */
template<class T> class Xlist {
    friend class XlistIterator<T>;
    
protected:
    /** 
    	Maximum value of list elements, that are deleted from the list put
     	still pointed to by some iterator.
    */
    enum { TABLESIZE = 10 };

    /**
       First element of this list.
    */
    XlistElement<T> *first;

    /** 
    	Last element of this list.
    */
    XlistElement<T> *last;

    /** 
    	If we have removed an element whose reference count is non-zero, 
	it is put in the toRemove table.
    */
    XlistElement<T> *toRemove[TABLESIZE];
    
    /** 
  	Iterator class listIterator calls this method, when
  	it releases given listElement.  	
  
  	Reduces the usage count of the element by one, and returns a pointer
 	to the next element.

 	If the element has already been removed (it's in the toRemove
 	table), it might get deleted: therefore this method must return
 	the next element.
 	@param *element Released listElement.
    */
    void release(XlistElement<T> *element);


    /**
       Marks that this list item is in use of an iterator. 
       This method is the opposite of method 
       release(listElement<T> *element).   
       @param *element listElement that is taken in use.
    */    
    void takeInUse(XlistElement<T> *element);
    
public:
    /**
       Constructor. 
    */ 
    Xlist();

    /**
       Adds given item to the front of this list.
       @param *item Item that will be added to the
       front of this list.
	
    */
    void prepend(T *item);

    /**
       Adds given item to the tail of this list.
       @param *item Item that will be added to 
       the tail of this list.
    */
    void append(T *item);

    /**
       Removes given item from this list. Removes the item
       pointed by argument from the list.	    
       @param
       @return
    */
    T* remove(T *item);

    /**
       Adds a given item to the front of this list.
       Method checks that item is not already 
       in this list.
       @param *item Item that will be added to front
       of this list.
    */
    void uniquePrepend(T *item);

    /**
       Adds a given item to the tail of this list.
       Method checks that item is not already 
       in this list.
       @param *item Item that will be added to tail
       of this list.
    */
    void uniqueAppend(T *item);

    /**
       Tests whether this list is empty or not.
       @return True if list is empty, otherwise false.
    */
    bool isEmpty();

    /**
       Finds if the given item already exists in this list.
       An empty item (=NULL) is never found from the 
       list. 
       @param 
       @return Pointer to the item, if it was found, 0 otherwise.
    */
    T *contains(T *item);

    /**
       Empties this list.
    */
    void deleteContents();

    /**
       In order to walk through this list, an iterator is needed    	
       @return An iterator to this list.
    */
    inline XlistIterator<T> giveIterator() {
	return XlistIterator<T>(this);
    };
    
    ~Xlist();
};




// class listIterator
//  This is the standard iterator for the list. It's quite like the
//  STL iterators.

/**
  Class listIterator. Iterator is needed in order to 
  overhaul instances of the class list. 
 */
template<class T>
class XlistIterator {
protected:
    /// List that this listIterator is pointing
    Xlist<T> *_list;
    
    /// Current element that this listIterator is pointing
    XlistElement<T> *_e;
	
public:
    /// Constructor.
    XlistIterator();
    
    /**
       Constructor. 
       @param *l List to which this listIterator will point to.
    */
    XlistIterator(Xlist<T> *l);

    /**
       Constructor.
       @param *l List to which this listIterator will point to.
       @param *el
    */
    XlistIterator(Xlist<T> *l, XlistIterator<T> *el);

    // Copy constructor
    XlistIterator(const XlistIterator& rhs);

    // State describing commands
    inline bool valid()   { return (_e != 0); }

    /**
       @return True if iterator points to the first item
       of the list, false otherwise.
    */
    inline bool atFirst() { return (_e == _list->first); }

    /**
       @return True if iterator points to the last item
       of the list, false otherwise.
    */
    inline bool atLast()  { return (_e == _list->last); }
    
    T* operator*() const;
    
    // Movement commands
    ///  Moves the iterator to the first element in this list     
    void first();    

    ///  Move the iterator to the last element in this list     
    void last();

/**
   The preincrement operator for iterators; move it one step forward.

   @param
   @return
*/
    XlistIterator<T>& operator++();

/**
   The postincrement operator for iterators; move iterator one step 
   forward. 
   @param
   @return
*/
    XlistIterator<T> operator++(int);


/**
   The predecrement operator for iterators; move iterator one step back.
   @param
   @return
*/
    XlistIterator<T>& operator--();

/**
   The postdecrement operator for iterator; move iterator one step back.
   @param
   @return
*/
    XlistIterator<T> operator--(int);
    
    // Assignment
/**
   The assignment operator for iterators.    

   This is needed, because we want to know which elements in the list
   are in use, so the compiler generated wont do.         
   @param
   @return
*/
    XlistIterator<T>& operator=(const XlistIterator<T>& rhs);

    // Change the current item
/**
   Change the content at this palce in the list.     

   Because appending and prepending to the list isn't always enough,
   this is a way to change the actual data in the list.           
   @param
   @return
*/
    T* changeContent(T *);

    /// Destructor
    ~XlistIterator();
};


//----------------------------------------
// listIterator implementation
//----------------------------------------

//---------------------------------------------------------------------
// listIterator::listIterator
//  The constructors.
//
// Copy constructor MUST be defined here, because we keep count on
// of the elements, and a compiler generated copy constructor wouldn't 
// do the thing right.
// --------------------------------------------------------------------

template<class T>
XlistIterator<T>::XlistIterator() : _list(0), _e(0) { }

template<class T>
XlistIterator<T>::XlistIterator(Xlist<T> *l) : _list(l), _e(0) {
    if (_list != 0) {
	_e = _list->first;
	_list->takeInUse(_e);
    }
}

template<class T>
XlistIterator<T>::XlistIterator(Xlist<T> *l, XlistIterator<T> *el) : _list(l), _e(0) {
    if (_list != 0) {
	_e = el;
	_list->takeInUse(_e);
    }
}

template<class T>
XlistIterator<T>::XlistIterator(const XlistIterator& rhs) : _list(rhs._list), _e(rhs._e) {
    _list->takeInUse(_e);
}


//---------------------------------------------------------------------
// listIterator<T>::operator*
//  Return the value of the item.
//
// --------------------------------------------------------------------


template<class T>
T*
XlistIterator<T>::operator*() const
{
    if (_e != 0)
	return _e->item;
    else
	return 0;
}


//---------------------------------------------------------------------
// listIterator<T>::first
//  Move the iterator to the first element in the list
//
// --------------------------------------------------------------------

template<class T>
void
XlistIterator<T>::first()
{
    if (_list != 0) {
	_list->release(_e);
	_e = _list->first;
	_list->takeInUse(_e);
    }
}


//---------------------------------------------------------------------
// listIterator<T>::last
//  Move the iterator to the last element in list
//
// --------------------------------------------------------------------

template<class T>
void
XlistIterator<T>::last()
{
    if (_list != 0) {
	_list->release(_e);
	_e = _list->last;
	_list->takeInUse(_e);
    }
}


//---------------------------------------------------------------------
// listIterator<T>::operator++
//  The preincrement operator for iterators; move it one step forward.
//
// --------------------------------------------------------------------

template<class T>
XlistIterator<T>&
XlistIterator<T>::operator++()
{
    if (_list != 0) {
	// Take the next now, because e might get deleted
	XlistElement<T> *next = 0;
	if (_e != 0)
	    next = _e->next;
	
	_list->release(_e);
	_e = next;
	_list->takeInUse(_e);
    }
    return *this;
}


//---------------------------------------------------------------------
// listIterator<T>::operator++
//  The postincrement operator for iterators; move iterator one step
//  forward. 
// 
// --------------------------------------------------------------------

template<class T>
XlistIterator<T>
XlistIterator<T>::operator++(int)
{
    XlistIterator<T> tmp = *this;
    ++*this;
    return tmp;
}


//---------------------------------------------------------------------
// listIterator<T>::operator--
//  The predecrement operator for iterators; move iterator one step
//  back. 
// 
// --------------------------------------------------------------------

template<class T>
XlistIterator<T>&
XlistIterator<T>::operator--()
{
    if (_list != 0) {
	// Take the previous now, because _e might get deleted
	XlistElement<T> *prev = 0;
	if (_e != 0)
	    prev = _e->prev;
	    
	_list->release(_e);
	_e = prev;
	_list->takeInUse(_e);
    }
    return *this;
}


//---------------------------------------------------------------------
// listIterator<T>::operator--
//  The postdecrement operator for iterator; move iterator one step
//  back. 
//
// --------------------------------------------------------------------

template<class T>
XlistIterator<T>
XlistIterator<T>::operator--(int)
{
    XlistIterator<T> tmp = *this;
    --*this;
    return tmp;
}


//---------------------------------------------------------------------
// listIterator<T>::operator=
//  The assignment operator for iterators.
//
// This is needed, because we want to know which elements in the list
// are in use, so the compiler generated wont do.
// --------------------------------------------------------------------

template<class T>
XlistIterator<T>&
XlistIterator<T>::operator=(const XlistIterator<T>& rhs)
{
    // check for assignment to self
    if (this == &rhs) return *this;
    if (_list == rhs._list && _e == rhs._e) return *this;
	
    if (_list != 0) 
	_list->release(_e);

    _list = rhs._list;
    _e = rhs._e;

    if (_list != 0) 
	_list->takeInUse(_e);
	
    return *this;
}


//---------------------------------------------------------------------
// listIterator::changeContent
//  Change the content at this palce in the list.
//
// Because appending and prepending to the list isn't always enough,
// this is a way to change the actual data in the list. 
// --------------------------------------------------------------------

template<class T>
T*
XlistIterator<T>::changeContent(T *item)
{
    T *oldItem = 0;
    if (_e != 0) {
	oldItem = _e->item;
	_e->item = item;
    }

    return oldItem;
}


//---------------------------------------------------------------------
// listIterator<T>::~listIterator
//  The destructor.
//
// It's essential to release the element in order to keep the counters 
// right.
// --------------------------------------------------------------------

template<class T>
XlistIterator<T>::~XlistIterator()
{
    if (_list != 0) 
	_list->release(_e);
}




//----------------------------------------
// list implementation
//----------------------------------------



//---------------------------------------------------------------------
// list::list
//  The constructor.
//
// The list is empty, when first and last equal null. The toRemove
// table is used to store elements, which have been deleted but are
// still pointed to by some iterator.
// --------------------------------------------------------------------

template<class T>
Xlist<T>::Xlist()
{
    first = last = 0;
#ifdef BASSIST_LIST_REFCOUNT
    for (int i=0; i < TABLESIZE; i++)
	toRemove[i] = 0;
#endif
}


//---------------------------------------------------------------------
// list::prepend
//  Adds the item in front of the list.
//
// Adds only a pointer to the item, does no copy.
// --------------------------------------------------------------------

template<class T>
void
Xlist<T>::prepend(T* item)
{
    XlistElement<T> *element = new XlistElement<T>;
    element->item = item;

    // Because we are adding to the front, last is affected only when
    // the list is empty 
    if (isEmpty())
	first = last = element;
    else {
	element->next = first;
	first = first->prev = element;
    }
}

//---------------------------------------------------------------------
// list::append
//  Adds the item as the last one on the list.
//
// Adds only a pointer to the item, does no copy.
// --------------------------------------------------------------------

template<class T>
void
Xlist<T>::append(T* item)
{
    XlistElement<T> *element = new XlistElement<T>;
    element->item = item;

    // Because we are adding to the front, first is affected only when
    // the list is empty 
    if (isEmpty())
	first = last = element;
    else {
	element->prev = last;
	last = last->next = element;
    }
}


//---------------------------------------------------------------------
// list::uniquePrepend
//  Adds the item in front of the list, if it isn't on the list
//  already. 
//
// Adds only a pointer to the item, does no copy.
// --------------------------------------------------------------------

template<class T>
void
Xlist<T>::uniquePrepend(T* item)
{
    if (this->contains(item) == 0)
	this->prepend(item);
}


//---------------------------------------------------------------------
// list::uniqueAppend
//  Adds the item as the last one on the list, if it isn't on the list
//  already. 
//
// Adds only a pointer to the item, does no copy.
// --------------------------------------------------------------------

template<class T>
void
Xlist<T>::uniqueAppend(T* item)
{
    if (this->contains(item) == 0)
	this->append(item);
}


//---------------------------------------------------------------------
// list::remove
//  Removes items from the list. 
//  
// Removes the item pointed by argument from the list structure, but
// does not deallocate the actual item. If the argument pointer is
// NULL, removes the first item on list.
//
// Returns pointer to the item removed from the list, or NULL, if list 
// was empty or element was not found.
// --------------------------------------------------------------------

template<class T>
T *
Xlist<T>::remove(T* item)
{
    if (isEmpty())
	return 0;

    // Start walkthrough
    XlistElement<T> *element = first;

    // Search for the item
    if (item != 0) {
	while (item != element->item && element != last)
	    element = element->next;

	if (item != element->item)	// Not found
	    return 0;
    }

    // Element points now to the right list element (first or the
    // chosen one), and oldelement to the previous element, or the
    // same, if it was the first one
    
    T* object = element->item;

    // Fix the links in the list

    if (first == element)
	first = element->next;
    else
	element->prev->next = element->next;
    
    if (last == element)
	last = element->prev;
    else
	element->next->prev = element->prev;

#ifdef BASSIST_LIST_REFCOUNT
    if (element->count > 0) {
	// Woops, this element is in the use of at least one
	// iterator. To mark it as a deletable, invert it's count and
	// put it in the "to remove" table
#ifdef BASSIST_DEBUG
	DEBUG('l', "In use list element removed\n");
#endif

	element->count = -(element->count);
	
	// First correct the situation, if someone(s) in the
	// "to remove" table would point to this element as it's next
	int i;
	for (i = 0; i < TABLESIZE; i++) 
	    if (toRemove[i] != 0) {
		if (toRemove[i]->next == element) {
#ifdef BASSIST_DEBUG
		    DEBUG('l', "Next value corrected in list table\n");
#endif
		    toRemove[i]->next = element->next;
		}

		if (toRemove[i]->prev == element) {
#ifdef BASSIST_DEBUG
		    DEBUG('l', "Prev value corrected in list table\n");
#endif
		    toRemove[i]->prev = element->prev;
		}
	    }

	// Search for an empty spot in the table
	for (i = 0; toRemove[i] != 0 && i < TABLESIZE; i++)
	    ;
	
	if (toRemove[i] != 0) {
	    // Crisis: empty spot was not found
	    printf("\nInternal list error: too many deleted items in use\n");
	    exit(1);
	}
	toRemove[i] = element;

    } else
#endif	
	delete element;

#ifdef BASSIST_LIST_REFCOUNT
#ifdef BASSIST_DEBUG
    if (debugIsEnabled('L')) {
	int i;
	for (i = 0; i < TABLESIZE; i++) {
	    DEBUG('L', "toRemove[%d]: ", i);
	    if (toRemove[i] != 0) 
		DEBUG('L', "*\n");
	    else
		DEBUG('L', "\n");
	}
    }
#endif
#endif
	
    return object;
}

//---------------------------------------------------------------------
// list::release
//  Reduce the usage count of the element by one, and return a pointer 
//  to the next element.
//
// If the element has already been removed (it's in the toRemove
// table), it might get deleted: therefore this function must return
// the next element. 
// --------------------------------------------------------------------


template<class T>
void
Xlist<T>::release(XlistElement<T> *element)
{
    if (element == 0)
	return;

#ifdef BASSIST_LIST_REFCOUNT
    if (element->count > 0) {
	element->count--;
	return;
    }

    if (element->count == 0) {
	printf("\nInternal list error: an item in use was marked unused\n");
	fflush(stdout);
	exit(1);
    }

    // This element is on the removal list, so it's count is
    // negated 
    element->count++;

    if (element->count == 0) {
	// Count has finally dropped to 0, remove the element
	// from the list
	int i;
	for (i = 0; toRemove[i] != element && i < TABLESIZE; i++)
	    ;
	if (toRemove[i] != element) {
	    // Critical: count was below zero, but the element 
	    // wasn't in the table
	    printf("Internal list error\n");
	    exit(1);
	}
#ifdef BASSIST_DEBUG
	DEBUG('l', "Removing element from table[%d]\n", i);
#endif
	toRemove[i] = 0;
	delete element;
    }
#endif
    
    return;
}


//---------------------------------------------------------------------
// list::takeInUse
//  Mark that this list item is in use of an iterator.
//
// This is the opposite of release().
// --------------------------------------------------------------------

template<class T>
void 
Xlist<T>::takeInUse(XlistElement<T> *element)
{
    if (element != 0) {
#ifdef BASSIST_LIST_REFCOUNT
	if (element->count < 0)
	    element->count--;
	else
	    element->count++;
#endif
    }
}


//---------------------------------------------------------------------
// list::contains
//  Finds if the given item already exists on list.
//
// Returns a pointer to the item if it was found, 0 if it wasn't. An
// empty item (NULL) is never found on the list.
// --------------------------------------------------------------------

template<class T>
T *
Xlist<T>::contains(T *item)
{
    if (isEmpty() || item == 0)
	return 0;

    XlistElement<T> *element = first;

    while (item != element->item && element != last) 
	element = element->next;

    if (item != element->item)	// Not found
	return 0;
    else
	return item;
}


//---------------------------------------------------------------------
// list::isEmpty
//  Tests whether the list is empty or not.
//
// Returns true if list is empty.
// --------------------------------------------------------------------

template<class T>
bool
Xlist<T>::isEmpty()
{
    // List is empty, if last and first are empty
    if (first == 0)
	return true;
    else
	return false;
}


//---------------------------------------------------------------------
// list::deleteContents
//  Empties the list.
//
// --------------------------------------------------------------------

template<class T>
void
Xlist<T>::deleteContents()
{
    while (remove(0) != 0)	// Deallocates until list is empty.
	;
}


//---------------------------------------------------------------------
// list::~list
//  The destructor.
//
// There is a potential problem with the iterators: when the list is
// destructed, they point to nowhere. 
// --------------------------------------------------------------------

template<class T>
Xlist<T>::~Xlist()
{
    deleteContents();

#ifdef BASSIST_LIST_REFCOUNT
    for (int i = 0; i < TABLESIZE; i++)
	if (toRemove[i] != 0)
	    delete toRemove[i];
#endif
}

#endif // LIST_H

//-----------------------------//
// Local Variables:            //
// mode: c++                   //
// make-backup-files: t        //
// End:                        //
//-----------------------------//
