// record.cc
//
// Created 1998 by Karri-Pekka Laakso
//
// $Id: record.cc,v 1.11 1999/08/17 18:41:35 jks Exp $
//

#include "record.h"
#include "main.h"
#include <stdio.h>

extern int yylineno;


//---------------------------------------------------------------------
// record::record
//  The record class ordinary constructor.
//
// 
// --------------------------------------------------------------------

record::record(const char *n, Xlist<variable> *f) : name(n), type("record_"),
    base(0), fields(f), recRefs(new Xlist<variable>), vars(new Xlist<variable>),
    ownVars(0), haveInherited(false)
{
    type += name;

    if (recordByName(n))
	c.parse_error("redefinition of record '%s'", n);
}

//---------------------------------------------------------------------
// record::record
//  The record class derived record constructor.
//
// 
// --------------------------------------------------------------------

record::record(const char *n, const char *b, Xlist<variable> *f) :
    name(n), type("record_"), base(recordByName(b)), fields(f), 
    recRefs(new Xlist<variable>), vars(new Xlist<variable>), 
    ownVars(0), haveInherited(false)
{
    type += name;

    if (recordByName(n))
	c.parse_error("redefinition of record '%s'", n);

    if (!base)
	c.parse_error("record '%s' derived from nonexistent record '%s'", n, b);
}

//---------------------------------------------------------------------
// record::recordByName
//  Find the record called 'name'.
//
// This function is used to avoid redefinition of existing records,
// and to find the base records when defining a derived record.
// --------------------------------------------------------------------

record *
record::recordByName(const string &name)
{
    XlistIterator<record> i = c.records->giveIterator();

    for ( ; i.valid(); ++i) 
	if (name.compare((*i)->giveName()) == 0) 
	    return *i;
    
    return 0;
}

//----------------------------------------------------------------------
// record::ownVariables()
//----------------------------------------------------------------------

Xlist<variable> *
record::ownVariables()
{
    if (ownVars != 0)		// have cached copy
	return ownVars;
    if (base == 0)		// not a derived record
	return vars;

    ownVars = new Xlist<variable>;
    XlistIterator<variable> v = vars->giveIterator();
    for ( ; v.valid(); v++) {
	XlistIterator<variable> b = base->vars->giveIterator();
	bool matched = false;
	const string &name = (*v)->giveName();

	for ( ; b.valid(); b++) {
	    if (name == (*b)->giveName()) {
		matched = true;
		break;
	    }
	}

	if (!matched)
	    ownVars->append(*v);
    }

    return ownVars;
}


//---------------------------------------------------------------------
// record::inheritFields
//  
//
// 
// --------------------------------------------------------------------

void
record::inheritFields()
{
    // Cause the function to be idempotent
    if (haveInherited)
	return;
    haveInherited = true;

    // Inheritance is only sensible on derived records
    if (!base)
	return;

    // Make sure the base has inherited
    base->inheritFields();

    // FIXME: this is O(mn), could be more intelligent
    // For all fields of the base record:
    XlistIterator<variable> i = base->fields->giveIterator();
    for (; i.valid(); i++) {
	bool foundMatch = false;
	XlistIterator<variable> j = fields->giveIterator();

	// See if we already have a field with that name. If not,
	// create a new variable.
	for (; j.valid(); j++) {
	    if ((*i)->giveName() == (*j)->giveName()) {
		foundMatch = true;
		break;
	    }
	}
	if (!foundMatch) {
	    fields->append((*i)->createInherited(this));
	}
    }
}

//---------------------------------------------------------------------
// record::findReferences
//  
//
// 
// --------------------------------------------------------------------

void
record::findReferences()
{
    XlistIterator<variable> i = fields->giveIterator();
    for (; i.valid(); ++i)
	(*i)->findReferences();
}


//---------------------------------------------------------------------
// record::printCode
//  The code generating function.
//
// 
// --------------------------------------------------------------------

void
record::printCode()
{
    if (base)
	fprintf(c.file, "record %s: %s {", name.c_str(), base->name.c_str());
    else
	fprintf(c.file, "record %s {", name.c_str());

    XlistIterator<variable> i = fields->giveIterator();
    for (; i.valid(); ++i) {
	(*i)->printCode();
	if (!i.atLast())
	    fprintf(c.file, ", ");
    }
    fprintf(c.file, "}");
}


//---------------------------------------------------------------------
// record::collectTypes
//  
//
// 
// --------------------------------------------------------------------

void
record::collectTypes(Xlist<variable> *c, Xlist<variable> *o,
		     Xlist<variable> *s)
{
    XlistIterator<variable> i = fields->giveIterator();
    for (; i.valid(); ++i) {
	(*i)->collectTypes(c, o, s);
	(*i)->collectRecFields(recRefs, vars);
    }
}


//----------------------------------------
// recordRef - record reference class
//----------------------------------------



//---------------------------------------------------------------------
// recordRef::recordRef
//  The record reference class constructors.
//
// 
// --------------------------------------------------------------------

recordRef::recordRef(const string &r) : 
    variable(r.c_str()), recName(r), dFile(c.datafile), line(yylineno), rec(0)
{
    type = "record_" + recName;
    // The reference check is done afterwards
}


recordRef::recordRef(const string &r, const string &a) : 
    variable(a.c_str()), recName(r), dFile(c.datafile), line(yylineno), rec(0)
{
    type = "record_" + recName;
    // The reference check is done afterwards
}

// recordRef::createInherited
// almost like a copy constructor -- FIXME: should it be?
// uses new to allocate the return value

variable *
recordRef::createInherited(record *r) const
{
    ASSERT(myRecord != 0);	// can only be called for record vars
    ASSERT(c.localRecord == 0); // should only get called after parsing
    c.localRecord = r;
    variable *v = new recordRef(recName, name);
    c.localRecord = 0;

    return v;
}


//---------------------------------------------------------------------
// recordRef::findReferences
//  Finds the references to other variables.
//
// Reference finding is done after the compilation to make
// cross-references possible. 
// --------------------------------------------------------------------

void
recordRef::findReferences()
{
    if (rec == 0) {
	// Search for the record
	XlistIterator<record> i = c.records->giveIterator();
	for ( ; i.valid(); ++i) 
	    if (recName.compare((*i)->giveName()) == 0) {
		rec = (*i);
		break;
	    }

	// If the record was not found, we define it to prevent further
	// errors. 
	if (rec == 0) {
	    rec = new record(recName.c_str(), new Xlist<variable>);
	    c.records->append(rec);

	    c.reference_error(dFile.c_str(), line, "record '%s' is undefined", recName.c_str());
	}
    }
}


//---------------------------------------------------------------------
// recordRef::nameSpace
//  
//
// 
// --------------------------------------------------------------------

Xlist<variable>* 
recordRef::nameSpace()
{
     if (rec == 0)
	 findReferences();
     
     return rec->allFields(); 
}


//---------------------------------------------------------------------
// recordRef::printCode
//  The code generating function.
//
// 
// --------------------------------------------------------------------

void
recordRef::printCode() const
{
    fprintf(c.file, "ref %s", rec->giveName().c_str());    
    if (rec->giveName().compare(name) != 0)
	fprintf(c.file, " as %s", name.c_str());
}


//---------------------------------------------------------------------
// recordRef::genPointer
//  
//
// 
// --------------------------------------------------------------------

void
recordRef::genPointer()
{
    fprintf(c.file, "record_%s *%s;", rec->giveName().c_str(), codename.c_str());
}


//---------------------------------------------------------------------
// recordRef::genRecName
//  
//
// 
// --------------------------------------------------------------------

void
recordRef::genRecName()
{
    fprintf(c.file, "%s", rec->giveName().c_str());
}


// Local Variables:    
// mode: c++           
// make-backup-files: t
// c-basic-offset: 4   
// case-replace: nil   
// End:                
