// record.h
//
// Created 1998 by Karri-Pekka Laakso
//
// $Id: record.h,v 1.10 1999/08/17 18:41:35 jks Exp $
//

#ifndef RECORD_H
#define RECORD_H

#include "list.h"
#include "variable.h"

/// A class for the records in the model.
class record : public parseTreeNode {
protected:
    /// The name of this record.
    string name;
    /// The type of this record.
    string type;
    /// The record from which this record is derived.
    record *base;
    /// The list of the definitions inside this record.
    Xlist<variable> *fields;
    /// The list of the record references defined inside this record.
    Xlist<variable> *recRefs;
    /// The list of the variables defined in this record or inherited.
    Xlist<variable> *vars;
    /**
     * The list of the variables defined in this record that 
     * do not exist in the base record.
     */
    Xlist<variable> *ownVars;
    /// Whether inheritFields has already been called.
    bool haveInherited;
    
public:
    /** The ordinary constructor.
	When a new record definition is created, it is checked that
	there doesn't already exist a record having the same name. If
	one exists, a parse error is reported.
	@param n The name of the record.
	@param f The list of the fields of the record.
	@see isRedefinition
	@see compiler::parse_error */
    record(const char *n, Xlist<variable> *f);

    /**
     * The constructor for derived records.
     * @param n The name of the record.
     * @param b The name of the base record.
     * @param f List of the overridden fields in this record.
     */
    record(const char *n, const char *b, Xlist<variable> *f);

    /** Gives the name of this record.
	@see name
	@return a constant reference to the name of the record */
    inline const string &giveName() const { return name; }
    /** Gives the C++ type name of this record.
	@see type
	@return a constant reference to the C++ type name  */
    inline const string &giveType() const { return type; }
    /// Gives the base record of this one, if this is a derived record
    inline record *giveBase() const { return base; }

    /** Finds a record with the given name.
	@see record::record
	@return A pointer to the record, or 0 if none is found. */
    static record *recordByName(const string &n);
    
    /// Gives a pointer to the list of the definitions of this record.
    inline Xlist<variable> *allFields() { return fields; }
    /// Gives a pointer to the list of the record references of this record.
    inline Xlist<variable> *references() { return recRefs; }
    /// Gives a pointer to the list of the variables of this record.
    inline Xlist<variable> *variables() { return vars; }
    /** 
     * Gives a pointer to the list of the non-inherited vars of this record.
     * The list is computed on demand.
     */
    Xlist<variable> *ownVariables();

    /**
     * Inherits fields from the possible base record.
     * This is called by compiler::FIXME after parsing for all records,
     * and the method itself ensures that the base record has inherited
     * from its base. Only the first call does any work; subsequent
     * calls have no effect. 
     */
    void inheritFields();

    /** Finds and links all the variable references to the actual
	variables. This function is needed in the end of the parse
	phase, when all the variable and record references are checked
	through. This function call is repeated recursively to 
	the fields of this record.
	@see compiler::parseInput
	@see variableRef::findReferences */
    void findReferences();

    /** Prints the model specification to the standard output.
	@see node::printCode
	@see compiler::printTree */
    void printCode();

    /** Collects the variables of the model into the type lists. The
        variable adds itself to the type list corresponding to its C++
	type. This function call is repeated recusively to the fields
	of this record. At the same time the fields of this record are
	collected by type into two lists: the list of record
	references and the list of variables. 
	@see compiler::collectVariables
	@see variable::collectRecFields
	@see fields
	@see recRefs
	@see vars */
    void collectTypes(Xlist<variable> *c, Xlist<variable> *o,
		      Xlist<variable> *s);
};


/// A class for record references.
class recordRef: public variable {
    /// The name of the referred record.
    string recName;
    /** The name of the model specification file in which this
	reference was used. Needed when reporting reference errors.
	@see findReferences
	@see compiler::reference_error */
    string dFile;
    /** The line number of the model specification file in which this
	reference was used. Needed when reporting reference errors.
	@see findReferences
	@see compiler::reference_error */
    int line;

    /// The referred record.
    record *rec;
    
public:  
    /** The constructor for the simple record references.
	@param r The name of the referred record. */      
    recordRef(const string &r);
    /** The constructor for the aliased record references.
	@param r The name of the referred record.
	@param a The name of the record reference. */
    recordRef(const string &r, const string &a);

    /// Creates an inherited variable in another record, allocates with new
    virtual variable *createInherited(record *r) const;


    /** Finds and links all the variable references to the actual
	variables. This function is needed in the end of the parse
	phase, when all the variable and record references are checked
	through. This function searches for the referred record, 
	which should be found; if it is not found, an error occurs.
	@see compiler::parseInput
	@see compiler::reference_error */
    void findReferences();
    
    /** Gives a pointer to the list of the variable definitions inside
        the referred record. 
        @return a pointer to the list
	@see variable::nameSpace */
    Xlist<variable>* nameSpace();

    /** Prints the model specification to the standard output.
	@see node::printCode
	@see compiler::printTree */
    void printCode() const;

    /** Collects the variables of the model into the type lists. The
        variable adds itself to the type list corresponding to its C++
	type. Because a record reference is not actually a variable of
	the model, it is not added to any of the lists.
	@see compiler::collectVariables */
    void collectTypes(Xlist<variable> *, Xlist<variable> *,
		      Xlist<variable> *) const {};
    
    /** Collects the variables defined inside a record to two lists, a
        list of variable definitions and a list of record
	references. This instance is added to the list of
	record references.
	@see record::collectTypes */
    void collectRecFields(Xlist<variable> *r, Xlist<variable> *)
    { r->append(this); }
    
    /// Generates the code that creates a new instance of this record reference. 
    void genCreate() { }
    /// Generates a C++ pointer. 
    void genPointer();
    /// Generates the name of the referred record.
    void genRecName();
};


#endif // RECORD_H

// Local Variables:    
// mode: c++           
// make-backup-files: t
// c-basic-offset: 4   
// case-replace: nil   
// End:                
