// scope.h
//
// Created 1998 by Karri-Pekka Laakso
//
// $Id: scope.h,v 1.5 1999/08/10 07:55:10 jks Exp $
//

#ifndef SCOPE_H
#define SCOPE_H

#include "list.h"
#include "record.h"
#include <string.h>


#define CG_MAX_SCOPES 1+10

/** A class for supporting FOREACH loops.

    The scope class supports the use of the FOREACH loops. When the
    parser finds the beginning of a new FOREACH loop, it creates a
    corresponding scope, which iterates through the list of the
    loop. The scope controls the code generation of the loop by
    passing code generation requests from the parser to the parse tree
    nodes. Before the lexer copies code to the result file, it
    asks the active scope for permission. Permission is denied only if
    the list of the active scope or an outer scope is empty.  */
class scope {
    /// The name of the loop variable.
    string name;
    /// The scope identifier. This scope is the id'th nested scope.  
    int id;

    enum { CODEBUFLEN = 30000 };
    
    /// The iterator for lists containing records.
    XlistIterator<record> *r;
    /// The iterator for lists containing variables.
    XlistIterator<variable> *v;
    
    /// The indicator of the print permission. True if printing is allowed.
    bool echo;
    /** Indicates if the FOREACH loop code should be saved to the
	codeBuffer. It will be saved only during the first pass
	through the loop, after which saveCode is set to false. */
    bool saveCode;
    /// The buffer in which the code of the loop is saved. 
    string codeBuffer;
    
    /** Tests if the list iterator points to a valid item, i.e.
	there is at least one item to process on the list.
	@return true if the iterator points to a valid item. */
    inline bool validIterator() { return (r == 0) ? v->valid() : r->valid(); }
    /// Moves the list iterator to point to the next item on the list.
    inline void nextVar() { if (r == 0)  ++(*v); else ++(*r); }
    
    /** Tests if this scope iterates a list of records.
	@return true if this is a scope that iterates records */
    inline bool isRecordScope() { return (r == 0) ? false : true; }

    /** Gives a pointer to the current variable. Note that the result
	is valid only for scopes that iterate variables.
	@return pointer to the current variable */ 
    inline variable *giveVariable() const { return *(*v); } 
    /** Gives a pointer to the current record. Note that the result
	is valid only for scopes that iterate records.
	@return pointer to the current record */ 
    inline record *giveRecord() const { return *(*r); }

    
    /** Sets this scope active. This is actually the last part of the
	constructor for normal scopes. */
    void setActive();

    /** The table of all the existing scopes. The 0th element of the
	table is always the 0-scope, after which the existing scopes are
	placed in creation order, the innermost loop last. */
    static scope *scopes[CG_MAX_SCOPES];
    /** The id of the last existing scope. Corresponds to the id of
	the scope of the innermost loop. */
    static int lastScope;
    
    /** Finds the correct scope for the loop variable.
	The table of scopes is searched from the end to the beginning
	in order to find the scope that has the matching loop variable 
	name.
	@param name The name of the loop variable.
	@return a pointer to the matching scope, or 0 if a matching
	scope was not found */
    static scope *findScope(const string &name);
    
public:
    /** The constructor for the 0-scope.  This constructor should be
	used only once to construct the 0-scope, which exists always
	during the code generation. If the 0-scope already exists, an
	error occurs. */
    scope();
    /** The constructor for normal scopes. Creates the scope and
	assigns it a list iterator. The name of the list is given in
	the argument {\em group}. Sets the created scope active with
	the setActive method, which is actually the final part of the
	constructor.
	
	@param var The name of the loop variable.
	@param group The list to be iterated.  */
    scope(const string &var, const string &group);

    /** Gives the scope identifier.
	@return the scope identifier */
    inline int giveId() const { return id; }
    /** Indicates if this scope enables the printing of code to the
	result file.
	@return true if printing is enabled */
    inline bool echoOn() const { return echo; }
    /** Gives a pointer to the buffer in which the code of this
	scope is saved.
	@return pointer to the code buffer */
    inline const string &giveBuffer() const { return codeBuffer; }
    
    /** Saves the code in the loop buffer.
	The code is saved only during the first pass of the
	corresponding FOREACH loop and never for the 0-scope.
	@param text The code to be saved.
	@param len The length of the code. */
    inline void saveLoopCode(const string &text, int len) {
	if (id > 0 && saveCode)
	    codeBuffer += text;
    }

    /** Implements the end-of-the-loop routines and tests, if the
	repetition of this FOREACH loop should be continued. When the
	lexer finds an ENDFOR in the code template, it calls this
	function. The loop code is now completely in the code buffer,
	and no more code will be saved. If the previous scope is still
	saving code, the contents of the code buffer of this scope are
	catenated to the code buffer of the previous scope. If there are
	items left on the list, the list iterator is moved forward,
	and the return value tells the lexer whether the loop
	iteration continues or not.
	@param text The end-of-the-loop text, i.e. ENDFOR.
	@return true if the loop iteration continues */
    bool loopEnd(const string &text);

    /** Replaces the code generation rule with the code fragments
	received from the parse tree nodes. At first, insertText
	searches the correct scope using findScope. Then the current
	parse tree node is asked to generate the code corresponding to 
	the code generation rule by invoking one of its code
	generation functions.
	@see findScope
	@param var The loop variable name.
	@param text The code generation rule. */
    void insertText(const string &var, const string &text); 

    /** Gives the previous scope in the scope table.
	Needed when destroying the current active scope. */
    scope *previous() const { return scopes[id-1]; }

    /// Destroy the current scope.
    ~scope();
};


#endif // SCOPE_H


// Local Variables:
// mode: c++
// c-basic-offset: 4
// case-replace: nil
// make-backup-files: t
// End:
